/*!\file m2-list.h
 * \brief 双向链表，基于 glib 提供的内存管理功能实现
 * \author Li Yanrui
 * \date 2008.04.14
 * \example test.c
 * 基于 M2_List 类型及其相关接口，实现点云数据的线性表存储，
 * 用于演示 M2 双向列表的用法。
 */
#ifndef  M2_LIST_H
#define  M2_LIST_H

#include "m2.h"

/*! 
 * \brief 将 void 类型定义为 M2_Object，更易于理解与使用
 */
typedef void M2_Object;

/*!
 * \brief M2 链表结点
 */
typedef struct tag_M2_List_Node {
	struct tag_M2_List_Node *prev;	/*!< 指向前结点的指针 */
	struct tag_M2_List_Node *next;	/*!< 指向后结点的指针 */
	M2_Object *data;	/*!< 指向数据结点的指针 */
} M2_List_Node;

/*!
 * \brief M2 链表结构体
 */
typedef struct {
	M2_List_Node *first;	/*!< 指向链表首结点的指针 */
	M2_List_Node *last;	/*!< 指向链表尾结点的指针 */
	int length;		/*!< 链表长度 */
} M2_List;

/*!
 * 传递给 m2_list_foreach () 的回调函数类型
 * \param object [I]：链表结点的数据对象指针
 * \param user_data [I]：用户传入的数据指针
 */
typedef void (*M2_List_Func) (M2_Object * object, M2_Object * user_data);

/*!
 * 创建链表
 * \return 指向链表结构体的指针
 */
M2_List *m2_list_create(void);

/*!
 * 释放链表内存，在调用该接口之前，应先释放链表各结点中引入的数据对象内存，否则会引起内存泄漏
 * \param list [I]: 目标链表
 */
M2_List *m2_list_delete(M2_List * list);

/*!
 * 在链表尾部添加数据对象。 
 * \param list [I]：目标链表
 * \param object [I]：待插入的数据对象，函数内部为其新建链表结点
 * \return 插入数据对象后的链表
 */
M2_List *m2_list_append(M2_List * list, M2_Object * object);

/*!
 * 在链表某一结点之前插入数据对象
 * \param list [I]: 目标链表
 * \param node [I]: 插入位置
 * \param object [I]: 待插入的数据对象，函数内部为其新建链表结点
 * \return 插入数据对象后的链表
 */
M2_List *m2_list_insert_before(M2_List * list,
			       M2_List_Node * node, M2_Object * object);

/*!
 * 在链表某一结点之后插入数据对象
 * \param list [I]: 目标链表
 * \param node [I]: 插入位置
 * \param object [I]: 待插入的数据对象，函数内部为其新建链表结点
 * \return 插入数据对象后的链表
 */
M2_List *m2_list_insert_after(M2_List * list,
			      M2_List_Node * node, M2_Object * object);

/*!
 * 删除链表的指定结点
 * \param list [I]: 目标链表
 * \param node [I]: 待删除结点
 * \return 删除结点后的链表
 */
M2_List *m2_list_remove_node(M2_List * list, M2_List_Node * node);

/*!
 * 获取链表首结点指针
 * \param list [I]: 目标链表
 * \return 链表首结点指针
 */
M2_List_Node *m2_list_get_first(M2_List * list);

/*!
 * 获取链表尾结点指针
 * \param list [I]: 目标链表
 * \return 链表尾结点指针
 */
M2_List_Node *m2_list_get_last(M2_List * list);

/*!
 * 调整当前链表结点指针，使之指向前一结点
 */
#define m2_list_prev(node) ((node) ? (((M2_List_Node *)(node))->prev): NULL)

/*!
 * 调整当前链表结点指针，使之指向后一结点
 */
#define m2_list_next(node) ((node) ? (((M2_List_Node *)(node))->next): NULL)

/*!
 * 遍历链表
 * \param list [I]: 目标链表
 * \param func [I]: 用于链表结点处理的回调函数
 * \param object [I]: 用户传入的自定义数据对象指针
 * \param direction [I]: 遍历方向，值为 0，表示由首结点向后遍历；
 * 值为 1，表示由尾结点向前遍历
 */
void m2_list_foreach(M2_List * list,
		M2_List_Func func, M2_Object * object, int direction);

#endif
